%{

#include "common.h"
#include "parseprg.h"

int yyparse();
void yyerror(char*);

class node_stmt;
node_stmt* program; // The result of parsing

//-----------------------------------------------------------------------------
// Define types for nodes in parse tree:

class node_stmt {
  public:
  virtual void eval()=0;
};

class node_intexpr {
  public:
  virtual int eval()=0;
};

// These are just the locations of the variabels:
typedef int*      intvar; 
typedef mac_addr* macvar;
typedef ip_addr*  ipvar;

//-----------------------------------------------------------------------------

// A constant of type T - we just declare constant as variabels
template<class T> class Constant {
  T a;
  public:
  Constant(T value) : a(value) {}
  operator T* () { return &a; }
};

// Declare the different constant types:
typedef Constant<ip_addr> ipConst;
typedef Constant<mac_addr> macConst;
typedef Constant<int> intConst;

//-----------------------------------------------------------------------------
// The symbol table:

class Symtable {
  const int maxsymbols=20;
  const int maxsymlen=20;
    
  public:
  Symtable() {  clear(); }  
  void clear() { symc=0; }
  
  union value {
    macvar mac_value;
    ipvar  ip_value;
    intvar int_value;
  };

  struct SymEntry {
    char   name[maxsymlen+1]; // Name of symbol
    ParseProgram::types type; // Type of symbol    
    value  value;             // Value of symbol
  };
  
  // Returns the symbol entry for a given symbol. Returns 0 on error:
  SymEntry* getSymbol(char* name);
  
  private:
  
  SymEntry symbols[maxsymbols]; // Symbol tabel
  value    values[maxsymbols];  // values for the symbol that are automatically
                                // created.
  int symc;                     // Number of symbols in table  
} symtabel;  

#include <stdio.h>

Symtable::SymEntry* Symtable::getSymbol(char* name) {
  // Return symbol if it exists:
  for(int a=0; a<symc; a++)
    if(!strcmp(name,symbols[a].name)) // Symbol found    
        return &symbols[a];

  // Create the symbol:    
  if(symc==maxsymbols) { yyerror("Too many variabels"); return 0; }  
  if(strlen(name)>maxsymlen) { yyerror("Variabel name too long"); return 0; }
  strcpy(symbols[symc].name,name);
  
  // Ask the user about the symbol:
  void* p=0;
  ParseProgram::types type=ParseProgram::instance->getSymbol(name,p);
  symbols[symc].type=type;
  if(p!=0) // User has supplied a location of variabel:
    symbols[symc].value.int_value=(int*)p;
  else { // Use a defualt location
    switch(type) {
      case ParseProgram::tINT: symbols[symc].value.int_value=*new intConst(0); break;
      case ParseProgram::tMAC: { mac_addr a; a.setNull(); symbols[symc].value.mac_value=*new macConst(a); break; }
      case ParseProgram::tIP:  { ip_addr a;  a.setNull(); symbols[symc].value.ip_value=*new ipConst(a); break; }
    }
  }
  
  return &symbols[symc++];
}

//-----------------------------------------------------------------------------
// Statement nodes:

class JoinStmt:public node_stmt {
  node_stmt &a, &b;
  virtual void eval() { a.eval(); b.eval(); }
  public:
  JoinStmt(node_stmt& A, node_stmt& B) : a(A), b(B) {}
};

class If:public node_stmt {
  node_intexpr &boole;
  node_stmt &stmts;  
  virtual void eval() { if(boole.eval()) stmts.eval(); }
  public:
  If(node_intexpr& A, node_stmt& B) : boole(A), stmts(B) {}
};

class IfElse:public node_stmt {
  node_intexpr &boole;
  node_stmt &stmt1,&stmt2;  
  virtual void eval() { if(boole.eval()) stmt1.eval(); else stmt2.eval(); }
  public:
  IfElse(node_intexpr& A, node_stmt& B, node_stmt& C) : boole(A), stmt1(B), stmt2(C) {}
};

class AssignInt:public node_stmt {
  intvar location;
  node_intexpr& assign;
  
  virtual void eval() { *location=assign.eval(); }
  
  public:
  AssignInt(intvar A, node_intexpr& B) : assign(B), location(A) {}
};

//------------------------------------------------------------------------------
// Simple integer operations:

class Less:public node_intexpr {
  node_intexpr &a,&b;
  
  virtual int eval() {
    return a.eval()<b.eval();
  }
  
  public:
  Less(node_intexpr& A, node_intexpr& B) : a(A), b(B) {}  
};

class LessEq:public node_intexpr {
  node_intexpr &a,&b;
  
  virtual int eval() {
    return a.eval()<=b.eval();
  }
  
  public:
  LessEq(node_intexpr& A, node_intexpr& B) : a(A), b(B) {}  
};

class And:public node_intexpr {
  node_intexpr &a,&b;
  virtual int eval() {
    return a.eval() && b.eval();
  }
  
  public:  
  And(node_intexpr& A, node_intexpr& B) : a(A), b(B) {}  
};

class Or:public node_intexpr {
  node_intexpr &a,&b;
  
  virtual int eval() {
    return a.eval() || b.eval();
  }
  
  public:
  Or(node_intexpr& A, node_intexpr& B) : a(A), b(B) {}  
};

class Add:public node_intexpr {
  node_intexpr &a,&b;
  virtual int eval() {
    return a.eval() + b.eval();
  }
  
  public:  
  Add(node_intexpr& A, node_intexpr& B) : a(A), b(B) {}  
};

class Not:public node_intexpr {
  node_intexpr &a;
  
  virtual int eval() {
    return !a.eval();
  }
  
  public:
  Not(node_intexpr& A) : a(A) {}
};

class IntEq:public node_intexpr {
  node_intexpr &a,&b;
  
  virtual int eval() {
    return a.eval()==b.eval();
  }
  
  public:
  IntEq(node_intexpr& A, node_intexpr& B) : a(A), b(B) {}  
};

class intVarAsExpr:public node_intexpr {
  intvar i;
  virtual int eval() { return *i; }
  public:
  intVarAsExpr(intvar a): i(a) {}
};

//------------------------------------------------------------------------------
// Comparing variabels. Note that integer variabels are compared as integer
// expressions since a variabel is also an expression.

template<class T> class Equal:public node_intexpr {
  T a,b;
  
  virtual int eval() {
    return *a==*b;
  }
  
  public:
  Equal(T A, T B) : a(A), b(B) {}  
};

typedef Equal<macvar> EqualMac;
typedef Equal<ipvar>  EqualIp;

//-----------------------------------------------------------------------------
// Counter list

class node_cntlist {
  const int MaxCounters=maxcounterc;
  
  intvar table[MaxCounters]; // Array of variabels
  int tablec;                // Elements in array
  
  friend class addMac;
  
  public:
  node_cntlist() { tablec=0; }
  
  int addNode(intvar a) { 
    if(tablec==MaxCounters) { yyerror("Too many counters"); return 1; }
    table[tablec++]=a;
    return 0;
  };
};

// An addMac statement:
class addMac:public node_stmt {
  node_cntlist& counters;
  node_intexpr& toadd;
  macvar mac;
  
  virtual void eval() {
    // Call user routine:
    ParseProgram::instance->macIncr(*mac,counters.tablec,counters.table,toadd.eval());
  }
  
  public:
  addMac(macvar Mac, node_cntlist& Lst,node_intexpr& toAdd) 
    : counters(Lst), toadd(toAdd), mac(Mac) {}
};

//-----------------------------------------------------------------------------
%}

%union {
  node_stmt*    nstmt;
  node_intexpr* nintexpr;
  node_cntlist* ncntlist;
  intvar        nintvar;
  macvar        nmacvar;
  ipvar         nipvar;
}

%type <nstmt>    stmt stmtlist program
%type <nintexpr> intexpr
%type <ncntlist> cntlist

%token <nintvar> INTVAR
%token <nmacvar> MACVAR
%token <nipvar>  IPVAR

%token IF ELSE AND OR LEQ GEQ EQ NEQ LE GE ADDEQ

%left OR
%left AND 
%left LE GE EQ LEQ GEQ NEQ
%left '+'
%left '!'

%%

program : stmtlist                                  { program=$1; }

stmtlist: stmtlist stmt                             { $$=new JoinStmt(*$1,*$2); }
        | stmt                                      { $$=$1; }
	
stmt    : '{' stmtlist '}'                          { $$=$2; }
        | IF '(' intexpr ')' stmt ELSE stmt         { $$=new IfElse(*$3,*$5,*$7); }
	| IF '(' intexpr ')' stmt                   { $$=new If(*$3,*$5);   }
	| MACVAR '[' cntlist ']' ADDEQ intexpr ';'  { $$=new addMac($1,*$3,*$6); }
       	| INTVAR  '=' intexpr ';'                   { $$=new AssignInt($1,*$3); }
		
intexpr : intexpr LE   intexpr { $$=new Less(*$1,*$3);   }
        | intexpr LEQ  intexpr { $$=new LessEq(*$1,*$3); }
	| intexpr GE   intexpr { $$=new Less(*$3,*$1);   }
	| intexpr GEQ  intexpr { $$=new LessEq(*$3,*$1); }
	| intexpr AND  intexpr { $$=new And(*$1,*$3);    }
	| intexpr OR   intexpr { $$=new Or(*$1,*$3);     }
 	| intexpr '+'  intexpr { $$=new Add(*$1,*$3);    }
	| '!' intexpr          { $$=new Not(*$2);        }
	| '(' intexpr ')'      { $$=$2; }
	| INTVAR               { $$=new intVarAsExpr($1); }
	| intexpr EQ intexpr   { $$=new IntEq(*$1,*$3);                }
	| intexpr NEQ intexpr  { $$=new Not(*(new IntEq(*$1,*$3)));    }
	| MACVAR EQ MACVAR     { $$=new EqualMac($1,$3);               }
	| MACVAR NEQ MACVAR    { $$=new Not(*(new EqualMac($1,$3)));   }
	| IPVAR EQ IPVAR       { $$=new EqualIp($1,$3);                }
	| IPVAR NEQ IPVAR      { $$=new Not(*(new EqualIp($1,$3)));    }

cntlist : INTVAR ',' cntlist   { $3->addNode($1); $$=$3; }
        | INTVAR               { $$=new node_cntlist(); $$->addNode($1); }  

%%

#include "lex.yy.c"

//------------------------------------------------------------------------------
// Define module interface:

#include <assert.h>

void yyerror(char *s) {
  snprintf(errmsg,sizeof(errmsg),"Error compiling program: %s",s);
}

ParseProgram* ParseProgram::instance=0;

ParseProgram::ParseProgram() {
  assert(instance==0); // Only one instance is allowed because of global
                       // variabels.
  instance=this;		      
  
  // Reset variabels:
  symtabel.clear();
}

ParseProgram::~ParseProgram() {
  instance=0;
}

char* ParseProgram::setProgram(char* c) {
  // Parse program
  yy_scan_string(c);
  yyerror("Parse error"); // default error message
  if(yyparse()) return errmsg;
  return 0;
}

void ParseProgram::execProgram() {
  program->eval();
}
